/* Copyright (c) 2014 Smart Vision. All Rights Reserved.
 *
 *
 */


#include <string.h>
#include <stdio.h>
#include "nordic_common.h"
#include "ble_l2cap.h"
#include "ble_srv_common.h"
#include "app_util.h"
#include "nrf_error.h"
#include "ble_gatt.h"
#include "app_error.h"
#include "ble_sbs.h"
#include "sbs_variables.h"
#include "BootData.h"
#include "simple_uart.h"

#include "app_trace.h"
#include "SPI.h"

#define SBS_LOG  app_trace_log  /**< Used for logging details. */
#define SBS_DUMP DumpData /**< Used for dumping octet information to get details of bond information etc. */

ble_gatts_char_handles_t notify_data_handles;
ble_gatts_char_handles_t write_handles;
			
	
uint8_t          		m_sbs_state;                   /**< Current communication state. */
uint16_t 		 		m_record_ndx;				    /**< Current notify record index. */


extern uint16_t 		gu16_StartingEventID;
extern uint16_t		 	gu16_TotalEventLogCount;
extern uint16_t		 	m_max_record;
extern uint8_t			write_data[WRITE_BUFFER_SIZE];
extern ble_manu_bitfield_t manu_bit;
extern uint8_t 			sbs_app_timer[MAX_APP_TIMER];


extern uint8_t 	notify_data_type;
extern void start_sensor_timer(void);
extern void button_enable(void);
extern void button_disable(void);
void idel_state(void);

/**@brief Connect event handler.
 *
 * @param[in]   p_sbs   	Smart Vision Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_connect(ble_sbs_t * p_sbs, ble_evt_t * p_ble_evt)
{
    p_sbs->conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
}


/**@brief Disconnect event handler.
 *
 * @param[in]   p_sbs   Smart Vision Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_disconnect(ble_sbs_t * p_sbs, ble_evt_t * p_ble_evt)
{
    UNUSED_PARAMETER(p_ble_evt);
	if(notify_data_type!=NOTIFY_IDEL)	
	{
		idel_state();	
		NVIC_SystemReset();		
	}
    p_sbs->conn_handle = BLE_CONN_HANDLE_INVALID;
}

/**@brief TX_COMPLETE event handler.
 *
 * @details Handles TX_COMPLETE events from the BLE stack.
 *
 * @param[in]   p_slat       Fit Five Notify Record Service structure.
 * @param[in]   p_ble_evt  	 Event received from the BLE stack.
 */
static void on_tx_complete(ble_sbs_t * p_sbs, ble_evt_t * p_ble_evt)
{
	#ifdef NOTIFICATION_DEBUG
		simple_uart_putstring("\r\ntx_cmplt\r\n");			
	#endif		
	switch(notify_data_type)
	{
		case TOTAL_EVENT_LOG_QUERY:
				report_all_notify_record(p_sbs);	
				break;
		case READ_AVAILABLE_EVENT:
		case GET_OLDEST_N_EVENT:
				report_eventlog_record(p_sbs);	
				break;			
	}
}

void ble_sbs_on_ble_evt(ble_sbs_t * p_sbs, ble_evt_t * p_ble_evt)
{
    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            on_connect(p_sbs, p_ble_evt);
            break;
            
        case BLE_GAP_EVT_DISCONNECTED:
            on_disconnect(p_sbs, p_ble_evt);
            break;
		
        case BLE_GATTS_EVT_WRITE:
            break;		
		
        case BLE_EVT_TX_COMPLETE:
            on_tx_complete(p_sbs, p_ble_evt);
            break;				
            
        default:
            break;
    }
}

/**@brief Add characteristic as a string 
 * @param[in]   p_sbs      	   Smart Vision Service structure.
 * @param[in]   uuid           UUID of characteristic to be added.
 * @param[in]   p_char_value   Initial value of characteristic to be added.
 * @param[in]   char_len       Length of initial value. This will also be the maximum value.
 * @param[out]  p_handles      Handles of new characteristic.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t sbs_char_add_string(ble_sbs_t *		p_sbs,
										uint16_t        uuid,
										uint8_t *       p_char_value,
										uint16_t        char_len,
                         ble_gatts_char_handles_t *      p_handles)
{
    ble_uuid_t          ble_uuid;
    ble_gatts_char_md_t char_md;
    ble_gatts_attr_t    attr_char_value;
    ble_gatts_attr_md_t attr_md;

    APP_ERROR_CHECK_BOOL(p_char_value != NULL);
    APP_ERROR_CHECK_BOOL(char_len > 0);
    
    // The ble_gatts_char_md_t structure uses bit fields. So we reset the memory to zero.
    memset(&char_md, 0, sizeof(char_md));

	if(uuid==BLE_UUID_WRITE)
	{
		char_md.char_props.write  = 1;	
	}
	else if(uuid==BLE_UUID_NOTIFY_DATA)
	{
		char_md.char_props.read  = 1;
		char_md.char_props.notify = 1;			
	}

    char_md.p_char_user_desc = NULL;
    char_md.p_char_pf        = NULL;
    char_md.p_user_desc_md   = NULL;
    char_md.p_cccd_md        = NULL;
    char_md.p_sccd_md        = NULL;

	ble_uuid.type = p_sbs->sbs_uuid_type;
	ble_uuid.uuid = uuid;		
		
    memset(&attr_md, 0, sizeof(attr_md));

	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);	
	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.write_perm);
		
    attr_md.vloc       = BLE_GATTS_VLOC_STACK;		
    attr_md.rd_auth    = 0;
    attr_md.wr_auth    = 0;
    attr_md.vlen       = 0;
    
    memset(&attr_char_value, 0, sizeof(attr_char_value));

    attr_char_value.p_uuid       = &ble_uuid;
    attr_char_value.p_attr_md    = &attr_md;
    attr_char_value.init_len     = char_len;
    attr_char_value.init_offs    = 0;
    attr_char_value.max_len      = char_len;
    attr_char_value.p_value      = p_char_value;

    return sd_ble_gatts_characteristic_add(p_sbs->service_handle, &char_md, &attr_char_value, p_handles);
}



uint32_t ble_sbs_init(ble_sbs_t * p_sbs, ble_sbs_init_t * p_sbs_init)
{
    uint32_t      err_code;
	ble_uuid_t    service_uuid;
	ble_uuid128_t sbs_base_uuid = SBS_UUID_BASE;

	// Initialize service structure
	p_sbs->evt_handler                 = p_sbs_init->evt_handler;
	p_sbs->conn_handle                 = BLE_CONN_HANDLE_INVALID;
		
	// Add service (Smart Vision device)
	err_code = sd_ble_uuid_vs_add(&sbs_base_uuid, &p_sbs->sbs_uuid_type);
	if (err_code != NRF_SUCCESS)
	{
		return err_code;
	}

	service_uuid.type = p_sbs->sbs_uuid_type;
	service_uuid.uuid = SMARTBEACON_UUID_SERVICE;

	err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY, &service_uuid, &p_sbs->service_handle);
	if (err_code != NRF_SUCCESS)
	{
		return err_code;
	}

	if (p_sbs_init->write_data.length > 0)
	{		
		err_code = sbs_char_add_string(p_sbs,
									BLE_UUID_WRITE,
									p_sbs_init->write_data.p_str, 
									p_sbs_init->write_data.length, 					
									&write_handles);					
		if (err_code != NRF_SUCCESS)
		{
			return err_code;
		}
	}

	if (p_sbs_init->notify_data.length > 0)
	{		
		err_code = sbs_char_add_string(p_sbs,
									BLE_UUID_NOTIFY_DATA,
									p_sbs_init->notify_data.p_str, 
									p_sbs_init->notify_data.length, 					
									&notify_data_handles);					
		if (err_code != NRF_SUCCESS)
		{
			return err_code;
		}
	}	
    return NRF_SUCCESS;
}

/**@brief data notification

 */
uint32_t ble_sbs_notify_data(ble_sbs_t * p_sbs, uint8_t *u8_buffer)
{
	uint32_t               err_code;
	ble_gatts_hvx_params_t hvx_params;
	uint16_t 			   len;

	len = 20;
	memset(&hvx_params, 0, sizeof(hvx_params));
	
	#ifdef BLE_DATA_DBG
		SBS_DUMP(u8_buffer,len);
	#endif	

	hvx_params.handle   = notify_data_handles.value_handle;
	hvx_params.type     = BLE_GATT_HVX_NOTIFICATION;
	hvx_params.offset   = 0;
	hvx_params.p_len    = &len;
	hvx_params.p_data   = u8_buffer;
	
	err_code = sd_ble_gatts_hvx(p_sbs->conn_handle, &hvx_params);
	return err_code;
}
/**@brief 
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
uint32_t aduit_trail_record_send(ble_sbs_t * p_sbs,uint16_t EventID)
{
	uint32_t               err_code;
	ble_gatts_hvx_params_t hvx_params;
	uint8_t  			   trail_rcd[20];		
	uint16_t 			   len;

	sbs_app_timer[DISCC_TIMER] = DISCC_TIME;
	memset(trail_rcd, 0, sizeof(trail_rcd));
//	ReadDatafromExternalFlash(&EventID,&trail_rcd[0]);

	len = 20 ;	

	memset(&hvx_params, 0, sizeof(hvx_params));
	
	#ifdef BLE_DATA_DBG
		SBS_DUMP(trail_rcd,len);
	#endif	
	
	hvx_params.handle   = notify_data_handles.value_handle;
	hvx_params.type     = BLE_GATT_HVX_NOTIFICATION;
	hvx_params.offset   = 0;
	hvx_params.p_len    = &len;
	hvx_params.p_data   = &trail_rcd[0];
	
	err_code = sd_ble_gatts_hvx(p_sbs->conn_handle, &hvx_params);
	#ifdef NOTIFICATION_DEBUG
		SBS_LOG("m_record_ndx =%d,\r\n",m_record_ndx);							
	#endif		
	if(err_code == NRF_SUCCESS)
	{
		if (m_sbs_state == STATE_SBS_ACTIVE)
		{
			m_record_ndx++;
		}
	}
	return err_code;
}
uint32_t report_eventlog_record(ble_sbs_t * p_sbs)
{
	uint32_t      err_code;	

    while(m_sbs_state == STATE_SBS_ACTIVE)
    {
		if (gu16_TotalEventLogCount==0)
		{
			idel_state();
			break;
		}
		else
		{
			err_code = aduit_trail_record_send(p_sbs,(gu16_StartingEventID));
			if( err_code != NRF_SUCCESS)
			{
				return err_code;
			}				
		}	
		switch(err_code)
		{
			case NRF_SUCCESS:
				if (m_sbs_state == STATE_SBS_ACTIVE)
				{
					gu16_StartingEventID++;
					if(gu16_StartingEventID > MAX_EVENT_RECORDS)
					{
						gu16_StartingEventID = 1;
					}
					gu16_TotalEventLogCount--;
				}
				break;
		
//			case BLE_ERROR_NO_TX_BUFFERS:
//				// Wait for TX_COMPLETE event to retry transmission
//				#ifdef NOTIFICATION_DEBUG
//					simple_uart_putstring("\r\nNO TX buf\r\n");			
//				#endif				
//				break;
			
            case NRF_ERROR_INVALID_STATE:
                // Notification is probably not enabled. Ignore request.
				idel_state();
                return err_code;
				
			default:
				// Report error to application
				if (p_sbs->error_handler != NULL)
				{
					p_sbs->error_handler(err_code);
				}

				// Make sure state machine returns to the default state
				idel_state();
				return err_code;						
		}
	}
    return NRF_SUCCESS;		
}


/**@brief 
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
uint32_t report_all_notify_record(ble_sbs_t * p_sbs)
{
	uint32_t      err_code;	
	
    while(m_sbs_state == STATE_SBS_ACTIVE) 
    {
		if (m_record_ndx > m_max_record)
		{
			idel_state();
		}
		else 
		{
			err_code = aduit_trail_record_send(p_sbs,m_record_ndx);
			if( err_code != NRF_SUCCESS)
			{
				return err_code;
			}
			switch(err_code)
			{
				case NRF_SUCCESS:
					// do nothing
					break;
			
//				case BLE_ERROR_NO_TX_BUFFERS:
//					// Wait for TX_COMPLETE event to retry transmission
//					#ifdef NOTIFICATION_DEBUG
//						simple_uart_putstring("\r\nNO TX buf\r\n");			
//					#endif					
//					break;
				
				case NRF_ERROR_INVALID_STATE:
					// Notification is probably not enabled. Ignore request.
					idel_state();
					return err_code;			
					
				default:
					// Report error to application
					if (p_sbs->error_handler != NULL)
					{
						p_sbs->error_handler(err_code);
					}
					// Make sure state machine returns to the default state
					idel_state();
					return err_code;						
			}			
		}	
	}
    return NRF_SUCCESS;	
}

/**@brief Set communication state.
 *
 * @param[in]   new_state   New communication state.
 */
void state_set_sbs(uint8_t new_state)
{
    m_sbs_state = new_state;
}

void idel_state(void)
{
	#ifdef NOTIFICATION_DEBUG
		simple_uart_putstring("\r\n ideal state\r\n");			
	#endif		
	state_set_sbs(STATE_NO_COMM);	
	notify_data_type = NOTIFY_IDEL;	
	spi_disable(SPI1);		
	#ifdef BLE_DEBUG
		simple_uart_putstring("Req Processed\r\n");			
	#endif		
}

